/*
  BiologicalCell.h

  Implementation of biological cell

  Copyright (C) 2005 Scott Christley

  Author: Scott Christley <schristl@nd.edu>
  Date: 2005
   
  This file is part of LimbChondrogenesis.
*/

#ifdef GNUSTEP
#include <Foundation/Foundation.h>
#else
#import <Cocoa/Cocoa.h>
#endif

#import "CellColony.h"

@interface BiologicalCell : NSObject
{
  CellColony *cellColony;
  unsigned int cellSize;
  unsigned int *spatialExtent[2];
  int cellType;
  int fibroProduction;
  int actMemoryProduce;
  BOOL stuckOnFibronectin;
  int activatorSum;
  int activatorThreshold;
}

- initWithColony: (CellColony *)aColony ofSize: (unsigned int)aSize
	     atX: (unsigned int)x atY: (unsigned int)y ofType: (int)aType;

- (void)buildCellAtX: (unsigned int)x atY: (unsigned int)y;
- (void)changeShape;
- (void)moveWithProbability: (double)aProb;
- (void)roundUp;

- (int)cellType;
- (void)setCellType: (int)aType;
- (unsigned int)cellSize;
- (unsigned int *)xSpatialExtent;
- (unsigned int *)ySpatialExtent;

- (int)fibroProduction;
- (void)incrementFibroProduction: (int)aNum;

- (void)recordActivatorProduce: (int)aNum;
- (int)activatorProduce;

- (int)activatorSum;
- (void)setActivatorSum: (int)aNum;

- (void)setStuckOnFibronectin: (BOOL)aFlag;
- (BOOL)stuckOnFibronectin;

- (int)activatorThreshold;
- (void)setActivatorThreshold: (int)mean withSD: (int)sd;

@end
