/*
  CellColony.h

  Implementation of a colony of biological cells.

  Copyright (C) 2005 Scott Christley

  Author: Scott Christley <schristl@nd.edu>
  Date: 2005
   
  This file is part of LimbChondrogenesis.
*/

#ifdef GNUSTEP
#include <Foundation/Foundation.h>
#else
#import <Cocoa/Cocoa.h>
#endif

#import "MultiScale2DGrid.h"

@interface CellColony : NSObject
{
  NSMutableArray *cells;
  NSDictionary *spatialGrid;
  int width;
  int height;

  MultiScale2DGrid *grid;
  NSDictionary *activatorGrid;
  NSDictionary *inhibitorGrid;
  int moleculeWidth;
  int moleculeHeight;
  int cellActivator;
  BOOL actMemory;

  NSDictionary *fibronectinGrid;
  int fibronectinWidth;
  int fibronectinHeight;
  int fibronectinScale;
}

- initWithSpatialGrid: (NSDictionary *)aGrid;

- (NSArray *)createColonyWithCellNumber: (unsigned int)count;
- (NSArray *)createColonyWithCellNumber: (unsigned int)count ofClass: (Class)cellClass;
- (NSDictionary *)spatialGrid;

- (void)recordActivatorMemory: (BOOL)aFlag;

- (NSArray *)randomizedCellList;
- (void)cellsPerform: (BOOL)randomize withInvocation: (NSInvocation *)anInvocation;

// cell diffusion
- (void)performDiffusion: (unsigned int)aCnt withShapeProbability: (double)sProb
     withMoveProbability: (double)mProb forCells: (NSArray *)cellList;
- (void)performDiffusion: (unsigned int)aCnt withShapeProbability: (double)sProb
     withMoveProbability: (double)mProb forCellType: (int)cellType;
- (void)performDiffusion: (unsigned int)aCnt withShapeProbability: (double)sProb
     withMoveProbability: (double)mProb forCellInvocation: (NSInvocation *)anInvocation;
- (void)performDiffusion: (unsigned int)aCnt withShapeProbability: (double)sProb
     withMoveProbability: (double)mProb;
- (void)performDiffusion: (unsigned int)aCnt withMoveProbability: (double)mProb;
- (void)checkStuckOnFibronectin: (NSDictionary *)aGrid;

// delegate methods for cell diffusion
- (void)setFibronectinGrid: (NSDictionary *)aGrid;
- (BOOL)validateMoveForCell: (id)aCell newXExtent: (unsigned int *)xExtent newYExtent: (unsigned int *)yExtent
cellSize: (int)cellSize;

// cell differentiation
- (BOOL)checkThreshold: (int)aValue forGrid: (NSDictionary *)aGrid;
- (BOOL)checkDoubleThreshold: (int)aValue forGrid: (NSDictionary *)aGrid;
- (BOOL)checkThreshold: (int)aValue withInvocation: (NSInvocation *)anInvocation;
//- (BOOL)checkDoubleThreshold: (int)aValue withInvocation: (NSInvocation *)anInvocation;
- (BOOL)checkThresholdForGrid: (NSDictionary *)aGrid;
- (void)cellRoundUp: (int)aValue forGrid: (NSDictionary *)aGrid;
- (void)cellDoubleRoundUp: (int)aValue forGrid: (NSDictionary *)aGrid;
- (void)productionForGrid: (NSDictionary *)aGrid;

// delegate methods for reaction-diffusion
- (void)setupReaction: (MultiScale2DGrid *)aGrid withActivatorGrid: (NSDictionary *)actGrid
    withInhibitorGrid: (NSDictionary *)inhibGrid;
- (BOOL)productionOfActivator: (int *)actProduce ofInhibitor: (int *)inhibProduce atX: (int)x atY: (int)y;

@end
