//
//  Diffusion.m
//  LimbChondrogenesis
//
//  Created by Scott Christley on 7/12/06.
//  Copyright 2006 __MyCompanyName__. All rights reserved.
//

#import "Diffusion.h"
#import "Grid2DArray.h"
#import "MultiScale2DGrid.h"

#import "internal.h"

@implementation Diffusion

- (void)performDiffusionForGrid: (NSDictionary *)aGrid
{
	// get matrix
	NSNumber *n = [aGrid objectForKey: @"height"];
	int height = [n intValue];
	n = [aGrid objectForKey: @"width"];
	int width = [n intValue];
	int (*theGrid)[height][width];
	Grid2DArray *m = [aGrid objectForKey: @"matrix"];
	theGrid = [m getMatrix];

	// diffusion parameters
	NSDictionary *params = [aGrid objectForKey: @"parameters"];
	NSString *s = [params objectForKey: @"diffusionRate"];
	double diffusionRate = [s doubleValue];
	s = [params objectForKey: @"diffusionTime"];
	int diffusionTime = [s intValue];

	s = [params objectForKey: @"alternate"];
	BOOL useAlt = NO;
	if (s) useAlt = [s boolValue];
	s = [params objectForKey: @"alternateDiffusionRate"];
	double altDiffusionRate = [s doubleValue];
	
	//printf("%lf %lf %d %lf %d\n", activatorDiffusion, activatorDiffusionOnFibronectin,
	// activatorDiffusionTime, inhibitorDiffusion, inhibitorDiffusionTime);
	
	int i, j, k, RND;
	int numTot, numRight, numLeft, numUp, numDown;
	double num, useDiffAll, useDiffOne;
	double diffRateAll, diffRateOne;
	double altDiffRateAll, altDiffRateOne;
	int matrix[height][width];
	int bCond;
	
	//printf("performDiffusion\n");

	// Assuming von Neumann neighborhood

	// equal rate of diffusion for each direction
	diffRateAll = diffusionRate;
	diffRateOne = diffusionRate / 4.0;
	// alternate diffusion rate
	altDiffRateAll = altDiffusionRate;
	altDiffRateOne = altDiffusionRate / 4.0;
	// boundary condition
	n = [aGrid objectForKey: @"boundaryCondition"];
	bCond = [n intValue];

	// perform diffusion some number of times
	for (k = 0; k < diffusionTime; ++k) {
			
		// Make copy of current grid
		for (i = 0; i < height; i++)
			for (j = 0; j < width; j++)
				matrix[i][j] = (*theGrid)[i][j];
			
		// diffuse
		for (i = 0; i < height; i++) {
			for (j = 0; j < width; j++) {
				
				// equal rate of diffusion for each direction
				useDiffAll = diffRateAll;
				useDiffOne = diffRateOne;
				// alternate diffusion rate on fibronectin
				if (useAlt) {
					//if (FN[i][j] > 0) {
					useDiffAll = altDiffRateAll;
					useDiffOne = altDiffRateOne;
					//}
				}
				
				if (matrix[i][j] > 0) {
					// determine amount
					num = (double)matrix[i][j];
					numTot = (int) floor((num * useDiffAll) + 0.5);
					//printf("%f %d\n", num, numTot);
					if (numTot > num)
						numTot = num;
					numRight = 0;
					numLeft = 0;
					numUp = 0;
					numDown = 0;
					
					if ((numTot == 0) || (num <= (4.0 / useDiffAll))) {
						//printf("num = %lf, matrix[%d][%d]=%d\n", num, i, j, matrix[i][j]);
						// its too small to take average
						// so randomize each one
						int k;
						
						for (k = 0; k < num; k++) {
							RND = (int) (RAND_NUM * (4.0 / useDiffAll));
							switch(RND) {
								case 0:
									++numRight;
									break;
								case 1:
									++numUp;
									break;
								case 2:
									++numLeft;
									break;
								case 3:
									++numDown;
									break;
							}
						}
					} else {
						// average the diffusion in all directions
						numRight = (int) floor((num * useDiffOne) + 0.5);
						numLeft = numRight;
						numUp = numRight;
						numDown = numRight;
						
						// rounding may cause the total to be
						// more than the actual number of molecules
						// that should be diffused
						if ((numRight * 4) > numTot) {
							do {
								// randomly decrease in one of
								// the directions until we reach
								// the right number of molecules
								RND = (int) (RAND_NUM * 4);
								switch(RND) {
									case 0:
										--numRight;
										if (numRight < 0) numRight = 0;
											break;
									case 1:
										--numLeft;
										if (numLeft < 0) numLeft = 0;
											break;
									case 2:
										--numUp;
										if (numUp < 0) numUp = 0;
											break;
									case 3:
										--numDown;
										if (numDown < 0) numDown = 0;
											break;
								}
							} while ((numRight + numLeft + numUp + numDown) > numTot);
						}
					}
					
					//printf("%d: numRight=%d, numUp=%d, numLeft=%d, numDown=%d\n", (*theGrid)[i][j],
					// numRight, numUp, numLeft, numDown);
					
					// diffuse the molecules
					// handle different boundary conditions
					(*theGrid)[i][j] -= numRight;
					if (j+1 == width) {
						// periodic can move but noflux does not
						if (bCond == PERIODIC_BOUNDARY)
							(*theGrid)[i][0] = (*theGrid)[i][0] + numRight;
						else
							(*theGrid)[i][j] += numRight;
					} else
						(*theGrid)[i][j+1] = (*theGrid)[i][j+1] + numRight;
					
					(*theGrid)[i][j] -= numLeft;
					if (j == 0) {
						// periodic can move but noflux does not
						if (bCond == PERIODIC_BOUNDARY)
							(*theGrid)[i][width-1] = (*theGrid)[i][width-1] + numLeft;
						else
							(*theGrid)[i][j] += numLeft;
					} else
						(*theGrid)[i][j-1] = (*theGrid)[i][j-1] + numLeft;
					
					(*theGrid)[i][j] -= numUp;
					if (i == 0) {
						// periodic can move but noflux does not
						if (bCond == PERIODIC_BOUNDARY)
							(*theGrid)[height-1][j] = (*theGrid)[height-1][j] + numUp;
						else
							(*theGrid)[i][j] += numUp;
					} else
						(*theGrid)[i-1][j] = (*theGrid)[i-1][j] + numUp;
					
					(*theGrid)[i][j] -= numDown;
					if (i+1 == height) {
						// periodic can move but noflux does not
						if (bCond == PERIODIC_BOUNDARY)
							(*theGrid)[0][j] = (*theGrid)[0][j] + numDown;
						else
							(*theGrid)[i][j] += numDown;
					} else
						(*theGrid)[i+1][j] = (*theGrid)[i+1][j] + numDown;
					
					//printf("%d\n", (*theGrid)[i][j]);
				}
			}
		}
	}
}

@end
