/*
  Grid2DArray.m

  Wrapper around 2D grid array.

  Copyright (C) 2005 Scott Christley

  Author: Scott Christley <schristl@nd.edu>
  Date: 2005
   
  This file is part of LimbChondrogenesis.
*/

#include <Foundation/Foundation.h>
#include "Grid2DArray.h"

#include "internal.h"

@implementation Grid2DArray

- initWithWidth: (unsigned int)aWidth andHeight: (unsigned int)aHeight andEncode: (NSString *)anEncode
{
  [super init];

  width = aWidth;
  height = aHeight;
  encode = anEncode;

  int i, j;
  if ([encode isEqual: idEncode]) {
    int size = sizeof(id[height][width]);
    matrix = malloc(size);
    //printf("id array [%d][%d] size: %d at: %p\n", height, width, size, matrix);
    id (*grid)[height][width] = matrix;
    for (i = 0;i < height; ++i)
      for (j = 0;j < width; ++j)
	(*grid)[i][j] = nil;
  } else if ([encode isEqual: intEncode]) {
    int size = sizeof(int[height][width]);
    matrix = malloc(size);
    //printf("int array [%d][%d] size: %d at: %p\n", height, width, size, matrix);
    int (*grid)[height][width] = matrix;
    for (i = 0;i < height; ++i)
      for (j = 0;j < width; ++j)
	(*grid)[i][j] = 0;
  } else if ([encode isEqual: longEncode]) {
    matrix = malloc(sizeof(long[height][width]));
    long (*grid)[height][width] = matrix;
    for (i = 0;i < height; ++i)
      for (j = 0;j < width; ++j)
	(*grid)[i][j] = 0;
  } else if ([encode isEqual: doubleEncode]) {
    matrix = malloc(sizeof(double[height][width]));
    double (*grid)[height][width] = matrix;
    for (i = 0;i < height; ++i)
      for (j = 0;j < width; ++j)
	(*grid)[i][j] = 0.0;
  } else if ([encode isEqual: boolEncode]) {
    matrix = malloc(sizeof(BOOL[height][width]));
    BOOL (*grid)[height][width] = matrix;
    for (i = 0;i < height; ++i)
      for (j = 0;j < width; ++j)
	(*grid)[i][j] = NO;
  } else {
    // throw exception or something
    printf("ERROR: Grid2DArray unknown encoding %s\n", [anEncode cString]);
  }

  return self;
}

- (void)dealloc
{
  free(matrix);
  [super dealloc];
}

- (unsigned int)getWidth { return width; }
- (unsigned int)getHeight { return height; }
- (void *)getMatrix { return matrix; }
- (NSString *)getEncode { return encode; }

- (void)clearMatrix
{
}

// count the number of connected components
- (void)visitNeighbors: (int)i : (int)j : (BOOL)periodic visitGrid: (void *)vGrid
{
	BOOL (*visitGrid)[height][width] = vGrid;
    int (*grid)[height][width] = matrix;

	// already visited, just return
	if ((*visitGrid)[i][j]) return;

	// indicate we visited this grid pixel
	(*visitGrid)[i][j] = YES;
	
	// nothing at this grid pixel, just return
	if (!(*grid)[i][j]) return;

	// check up
	if (i == 0) {
		if (periodic) [self visitNeighbors: height - 1 : j : periodic visitGrid: visitGrid];
	} else {
		[self visitNeighbors: i - 1 : j : periodic visitGrid: visitGrid];
	}

	// check down
	if (i == (height - 1)) {
		if (periodic) [self visitNeighbors: 0 : j : periodic visitGrid: visitGrid];
	} else {
		[self visitNeighbors: i + 1 : j : periodic visitGrid: visitGrid];
	}
	
	// check left
	if (j == 0) {
		if (periodic) [self visitNeighbors: i : width - 1 : periodic visitGrid: visitGrid];
	} else {
		[self visitNeighbors: i : j - 1 : periodic visitGrid: visitGrid];
	}
	
	// check right
	if (j == (width - 1)) {
		if (periodic) [self visitNeighbors: i : 0 : periodic visitGrid: visitGrid];
	} else {
		[self visitNeighbors: i : j + 1 : periodic visitGrid: visitGrid];
	}
}

- (int)countClusters: (BOOL)periodic
{
	int numClusters = 0;
	int i, j;
    int (*grid)[height][width] = matrix;

	// indicator grid for which points we have visited
	BOOL visitGrid[height][width];
	for (i = 0;i < height; ++i)
		for (j = 0;j < width; ++j)
			visitGrid[i][j] = NO;

	for (i = 0;i < height; ++i) {
		for (j = 0;j < width; ++j) {
			if ( (!visitGrid[i][j]) && ((*grid)[i][j]) ) {
				++numClusters;
				[self visitNeighbors: i : j : periodic visitGrid: visitGrid];
			}
		}
	}

	return numClusters;
}

@end
