/*
  ReactionDiffusion.h

  Implementation of reaction-diffusion chemical process.

  Copyright (C) 2005 Scott Christley

  Author: Scott Christley <schristl@nd.edu>
  Date: 2005
   
  This file is part of LimbChondrogenesis.
*/

@interface ReactionDiffusion : NSObject
{
  MultiScale2DGrid *grid;
  NSString *activatorName;
  NSString *inhibitorName;
  NSString *flagName;
  id delegate;
  BOOL delegatePreReaction;
  BOOL delegateProduction;
  BOOL delegateAfterCalculate;
  BOOL delegatePostReaction;
}

- initForGrid: (MultiScale2DGrid *)aGrid withActivatorGrid: (NSString *)aName
withInhibitorGrid: (NSString *)iName;

- (void)performReaction;
- (void)performDiffusion;

- (void)performDoubleReaction;
- (void)performDoubleDiffusion;

- (void)performReactionDiffusion;

@end

// delegate methods
@protocol ReactionDelegate
- (void)setupReaction: (MultiScale2DGrid *)aGrid withActivatorGrid: (NSDictionary *)actGrid
	withInhibitorGrid: (NSDictionary *)inhibGrid;
- (BOOL)preReaction;
- (BOOL)productionOfActivator: (int *)actProduce ofInhibitor: (int *)inhibProduce atX: (int)x atY: (int)y;
- (BOOL)doubleProductionOfActivator: (double *)actProduce ofInhibitor: (double *)inhibProduce atX: (int)x atY: (int)y;
- (void)afterCalculate;
- (void)postReaction;
@end
