//
//  ReactionDiffusionModel.m
//  LimbChondrogenesis
//
//  Created by Scott Christley on 7/10/06.
//  Copyright 2006 __MyCompanyName__. All rights reserved.
//

#import "ReactionDiffusionModel.h"
#import "MultiScale2DGrid.h"
#import "CellColony.h"
#import "Grid2DArray.h"
#import "ReactionDiffusion.h"
#import "BiologicalCell.h"

#import "internal.h"

@implementation ReactionDiffusionModel

- (void)doSim: (int)run
{
	// file pointer
	FILE *fp_cell, *fp_morpha, *fp_morphb, *fp_FN, *fp_FNcell, *fp_params;
	NSString *s;
	
	// output file name prefix
	NSMutableString *fileName = [NSMutableString string];
	[fileName appendString: [simParameters objectForKey: @"outputName"]];
	[fileName appendFormat: @"_%d", run];
	
	// open the output files
	NSString *outputFile = [NSString stringWithFormat: @"%@_cell.dat", fileName];
	fp_cell = fopen([outputFile cString], "wb+");
	if (fp_cell == NULL) {
		printf("cannot open file\n");
		exit(0);
	}
	
	outputFile = [NSString stringWithFormat: @"%@_morpha.dat", fileName];
	fp_morpha = fopen([outputFile cString], "wb+");
	if (fp_morpha == NULL) {
		printf("cannot open file\n");
		exit(0);
	}
	
	outputFile = [NSString stringWithFormat: @"%@_morphb.dat", fileName];
	fp_morphb = fopen([outputFile cString], "wb+");
	if (fp_morphb == NULL) {
		printf("cannot open file\n");
		exit(0);
	}
	
	outputFile = [NSString stringWithFormat: @"%@_FN.dat", fileName];
	fp_FN = fopen([outputFile cString], "wb+");
	if (fp_FN == NULL) {
		printf("cannot open file\n");
		exit(0);
	}
	
	outputFile = [NSString stringWithFormat: @"%@_FNcell.dat", fileName];
	fp_FNcell = fopen([outputFile cString], "wb+");
	if (fp_FNcell == NULL) {
		printf("cannot open file\n");
		exit(0);
	}
	
	outputFile = [NSString stringWithFormat: @"%@_matlab.dat", fileName];
	fp_params = fopen([outputFile cString], "wb+");
	if (fp_params == NULL) {
		printf("cannot open file\n");
		exit(0);
	}
	
	// save parameter set
	outputFile = [NSString stringWithFormat: @"%@_params.dat", fileName];
	[simParameters writeToFile: outputFile atomically: NO];
	
	// Initialize RNG
	s = [simParameters objectForKey: @"randomSeed"];
	srand([s intValue]);
	
	// create spatial grid
	s = [simParameters objectForKey: @"width"];
	int width = [s intValue];
	s = [simParameters objectForKey: @"height"];
	int height = [s intValue];
	s = [simParameters objectForKey: @"ratio"];
	int ratio = [s intValue];
	s = [simParameters objectForKey: @"boundaryCondition"];
	int bCond = PERIODIC_BOUNDARY;
	if ([s isEqualToString: @"NoFlux"])
		bCond = NOFLUX_BOUNDARY;
	MultiScale2DGrid *grid = [[MultiScale2DGrid alloc] initWithWidth: width andHeight: height];
	id cells = [grid newGridWithName: @"cells" atScale: 1 withEncode: idEncode];
	[cells setObject: [[NSNumber alloc] initWithInt: bCond] forKey: @"boundaryCondition"];

	id activator = [grid newGridWithName: @"activator" atScale: ratio withEncode: intEncode];
	//id activator = [grid newGridWithName: @"activator" atScale: ratio withEncode: doubleEncode];
	[activator setObject: [[NSNumber alloc] initWithInt: bCond] forKey: @"boundaryCondition"];
	id inhibitor = [grid newGridWithName: @"inhibitor" atScale: ratio withEncode: intEncode];
	//id inhibitor = [grid newGridWithName: @"inhibitor" atScale: ratio withEncode: doubleEncode];
	[inhibitor setObject: [[NSNumber alloc] initWithInt: bCond] forKey: @"boundaryCondition"];

	id fibronectin = [grid newGridWithName: @"fibronectin" atScale: ratio withEncode: intEncode];
	//id fibronectin = [grid newGridWithName: @"fibronectin" atScale: ratio withEncode: doubleEncode];
	[fibronectin setObject: [[NSNumber alloc] initWithInt: bCond] forKey: @"boundaryCondition"];
	id junk = [grid newGridWithName: @"junk" atScale: ratio withEncode: intEncode];
	[junk setObject: [[NSNumber alloc] initWithInt: bCond] forKey: @"boundaryCondition"];
	
	// test diffusion
	//Grid2DArray *m = [activator objectForKey: @"matrix"];
	//double (*activatorGrid)[height*ratio][width*ratio] = [m getMatrix];
	//(*activatorGrid)[60][60] = 10000;
	//m = [inhibitor objectForKey: @"matrix"];
	//double (*inhibitorGrid)[height*ratio][width*ratio] = [m getMatrix];
	//(*inhibitorGrid)[60][60] = 10000;
	
	// create cell colony
	s = [simParameters objectForKey: @"cells"];
	int numCells = [s intValue];
	id colony = [[CellColony alloc] initWithSpatialGrid: cells];
	NSArray *bc = [colony createColonyWithCellNumber: numCells];
	BOOL useMemory = NO;
	s = [simParameters objectForKey: @"useMemory"];
	if (s)
		useMemory = [s boolValue];
	[colony recordActivatorMemory: NO];
	s = [simParameters objectForKey: @"memoryDelay"];
	int memoryDelay = [s intValue];
	s = [simParameters objectForKey: @"memoryActive"];
	int memoryActive = [s intValue];
	
	[colony setFibronectinGrid: fibronectin];
	//NSArray *bc = [colony createColonyWithCellNumber: 100];
	//printf("%d\n", [bc count]);
	
	// reaction-diffusion process
	id reactDiff = [[ReactionDiffusion alloc] initForGrid: grid withActivatorGrid: @"activator"
										withInhibitorGrid: @"inhibitor"];
	[reactDiff setDelegate: colony];
	
	// initial uniform and random cell distribution
	[grid dumpGrid: @"cells" toFile: fp_cell];
	s = [simParameters objectForKey: @"initialCellDiffusionSteps"];
	if (s == nil)
		[colony performDiffusion: 100 withMoveProbability: 0.8];
	else
		[colony performDiffusion: [s intValue] withMoveProbability: 0.8];
	[grid dumpGrid: @"cells" toFile: fp_cell];
	
	// simulation parameters
	s = [simParameters objectForKey: @"cycles"];
	int totCycles = [s intValue];
	s = [simParameters objectForKey: @"cellMovement"];
	int cellMovement = [s intValue];
	s = [simParameters objectForKey: @"cellKeepShape"];
	double cellKeepShape = [s doubleValue];
	s = [simParameters objectForKey: @"cellRestProb"];
	double cellRestProb = [s doubleValue];
	BOOL speedyCells = NO;
	s = [simParameters objectForKey: @"speedyCells"];
	if (s) speedyCells = [s boolValue];
	s = [simParameters objectForKey: @"outputRate"];
	int outputRate = [s intValue];
	s = [simParameters objectForKey: @"activatorThreshold"];
	int activatorThreshold = [s intValue];
	s = [simParameters objectForKey: @"activatorThresholdDelay"];
	int activatorThresholdDelay = [s intValue];
	s = [simParameters objectForKey: @"fibronectinThreshold"];
  BOOL activatorThresholdVary = NO;
  s = [simParameters objectForKey: @"activatorThresholdVary"];
  if (s) activatorThresholdVary = [s boolValue];
  int activatorThresholdSD = 0;
  s = [simParameters objectForKey: @"activatorThresholdSD"];
  if (s) activatorThresholdSD = [s intValue];
	int fibronectinThreshold = [s intValue];
	s = [simParameters objectForKey: @"maxFibronectin"];
	int maxFibronectin = [s intValue];
	s = [simParameters objectForKey: @"fibronectinDelay"];
	int fibronectinDelay = [s intValue];
	s = [simParameters objectForKey: @"memoryThreshold"];
	int memoryThreshold = [s intValue];
	s = [simParameters objectForKey: @"fibronectinCellMovement"];
	int fibronectinCellMovement = [s intValue];
	s = [simParameters objectForKey: @"fibronectinCellKeepShape"];
	double fibronectinCellKeepShape = [s doubleValue];
	s = [simParameters objectForKey: @"fibronectinCellRestProb"];
	double fibronectinCellRestProb = [s doubleValue];
	s = [simParameters objectForKey: @"disableCellActivator"];
	BOOL disableCellActivator = [s boolValue];
	s = [simParameters objectForKey: @"disableCellActivatorThreshold"];
	int disableCellActivatorThreshold = [s intValue];
	s = [simParameters objectForKey: @"disableReactionDiffusion"];
	int disableReactionDiffusion = [s intValue];
	s = [simParameters objectForKey: @"enableMatlab"];
	BOOL enableMatlab = [s boolValue];
	
	// save parameters for Matlab
	fwrite(&width, sizeof(int), 1, fp_params);
	fwrite(&height, sizeof(int), 1, fp_params);
	fwrite(&ratio, sizeof(int), 1, fp_params);
	fwrite(&outputRate, sizeof(int), 1, fp_params);
	fwrite(&totCycles, sizeof(int), 1, fp_params);
	fwrite(&activatorThreshold, sizeof(int), 1, fp_params);
	fwrite(&fibronectinThreshold, sizeof(int), 1, fp_params);
	s = [simParameters objectForKey: @"activatorDiffusion"];
	double tmp1 = [s doubleValue];
	s = [simParameters objectForKey: @"activatorDiffusionTime"];
	double tmp2 = [s doubleValue];
	tmp1 = tmp1 * tmp2;
	fwrite(&tmp1, sizeof(double), 1, fp_params);
	s = [simParameters objectForKey: @"activatorOnActivator"];
	tmp1 = [s doubleValue];
	fwrite(&tmp1, sizeof(double), 1, fp_params);
	s = [simParameters objectForKey: @"activatorOnInhibitor"];
	tmp1 = [s doubleValue];
	fwrite(&tmp1, sizeof(double), 1, fp_params);
	s = [simParameters objectForKey: @"maxActivatorProduce"];
	int tmp3 = [s intValue];
	fwrite(&tmp3, sizeof(int), 1, fp_params);
	s = [simParameters objectForKey: @"cellActivator"];
	tmp3 = [s intValue];
	fwrite(&tmp3, sizeof(int), 1, fp_params);
	s = [simParameters objectForKey: @"inhibitorDiffusion"];
	tmp1 = [s doubleValue];
	s = [simParameters objectForKey: @"inhibitorDiffusionTime"];
	tmp2 = [s doubleValue];
	tmp1 = tmp1 * tmp2;
	fwrite(&tmp1, sizeof(double), 1, fp_params);
	s = [simParameters objectForKey: @"inhibitorOnActivator"];
	tmp1 = [s doubleValue];
	fwrite(&tmp1, sizeof(double), 1, fp_params);
	s = [simParameters objectForKey: @"inhibitorOnInhibitor"];
	tmp1 = [s doubleValue];
	fwrite(&tmp1, sizeof(double), 1, fp_params);
	s = [simParameters objectForKey: @"maxInhibitorProduce"];
	tmp3 = [s intValue];
	fwrite(&tmp3, sizeof(int), 1, fp_params);
	int itmp = (int)useMemory;
	fwrite(&itmp, sizeof(int), 1, fp_params);
	if (useMemory)
		fwrite(&memoryThreshold, sizeof(int), 1, fp_params);
	else
		fwrite(&activatorThreshold, sizeof(int), 1, fp_params);
	fwrite(&fibronectinThreshold, sizeof(int), 1, fp_params);
	fwrite(&maxFibronectin, sizeof(int), 1, fp_params);
	fwrite(&fibronectinDelay, sizeof(int), 1, fp_params);
	fclose(fp_params);

  // individual thresholds for each cell
  if (activatorThresholdVary) {
    int i;
    for (i = 0; i < [bc count]; ++i) {
      BiologicalCell *aCell = [bc objectAtIndex: i];
      [aCell setActivatorThreshold: activatorThreshold withSD: activatorThresholdSD];
    }
  }

	// setup invocations
	NSMethodSignature *sig = [[BiologicalCell new] methodSignatureForSelector: @selector(activatorProduce)];
	[sig retain];
	NSInvocation *actMemory = [NSInvocation invocationWithMethodSignature: sig];
	[actMemory setSelector: @selector(activatorProduce)];
	[actMemory retain];
	sig = [[BiologicalCell new] methodSignatureForSelector: @selector(stuckOnFibronectin)];
	[sig retain];
	NSInvocation *stuckInv = [NSInvocation invocationWithMethodSignature: sig];
	[stuckInv setSelector: @selector(stuckOnFibronectin)];
	[stuckInv retain];
	
	// main simulation loop
	int cycle;
	BOOL fibroProduction = NO;
	BOOL performRD = YES;
	BOOL triggerMemory = YES;
	BOOL activeMemory = YES;
	for (cycle = 0; cycle < totCycles; cycle++) {
		if ((cycle % outputRate) == 0) printf("cycle=%d\n",cycle);
		//printf("COUNT_FNcells=%d\n",COUNT_FNcells);
		
		// data output
		if ((cycle % outputRate) == 0) {
			[grid dumpGrid: @"cells" toFile: fp_cell];
			[grid dumpGrid: @"activator" toFile: fp_morpha];
			[grid dumpGrid: @"inhibitor" toFile: fp_morphb];
			[grid dumpGrid: @"fibronectin" toFile: fp_FN];
			//[grid dumpGrid: @"cells" toFile: fp_FNcell];
			[grid dumpGrid: @"cells" toFile: fp_FNcell withInvocation: actMemory];
		}

		// only test diffusion
		//[reactDiff performDiffusion];
		//[reactDiff performDoubleDiffusion];

		//
		if (triggerMemory) {
			if ((useMemory) && (cycle >= memoryDelay)) {
				[colony recordActivatorMemory: YES];
				triggerMemory = NO;
			}
		} else if (activeMemory) {
			if ((useMemory) && (cycle >= (memoryDelay + memoryActive))) {
				[colony recordActivatorMemory: NO];
				activeMemory = NO;
			}
		}

		// cell diffusion
		if (((cycle % cellMovement) == 0) || (speedyCells && !performRD)) {
			[colony performDiffusion: 1 withShapeProbability: cellKeepShape
				 withMoveProbability: (1 - cellRestProb)];
		}
		
		// cell diffusion on fibronectin
		if (((cycle % fibronectinCellMovement) == 0) || (speedyCells && !performRD)) {
			[colony performDiffusion: 1 withShapeProbability: cellKeepShape
				 withMoveProbability: (1 - cellRestProb) forCellInvocation: stuckInv];
		}

		// check if disable reaction diffusion
		if ((disableReactionDiffusion != 0) && (cycle > disableReactionDiffusion))
			performRD = NO;

		// reaction-diffusion
		if (performRD) {
			[reactDiff performReaction];
			[reactDiff performDiffusion];
			//[reactDiff performDoubleReaction];
			//[reactDiff performDoubleDiffusion];
//			int q;
//			for (q = 0; q < 10; ++q)
//				[reactDiff performReactionDiffusion];
		}

#if 1
		// fibronectin production
		if (useMemory)
			fibroProduction |= [colony checkThreshold: memoryThreshold withInvocation: actMemory];
		else if (cycle >= activatorThresholdDelay) {
      if (activatorThresholdVary)
        fibroProduction |= [colony checkThresholdForGrid: activator];
      else
        fibroProduction |= [colony checkThreshold: activatorThreshold forGrid: activator];
    }
    
		if (fibroProduction) {
			// produce fibronectin
			[colony productionForGrid: fibronectin];
			
			// check if cells get stuck on fibronectin
			[colony checkStuckOnFibronectin: fibronectin];
			
			// round up of cells on fibronectin
			[colony cellRoundUp: fibronectinThreshold forGrid: fibronectin];
			
			// once fibronectin production has started
			// determine if we should stop the reaction-diffusion
			if (disableCellActivator) {
				int max = [grid maxValueForIntGrid: @"activator"];
				if (max < disableCellActivatorThreshold)
					performRD = NO;
			}
		}
#endif
#if 0
		// fibronectin production
		if (useMemory)
			fibroProduction |= [colony checkDoubleThreshold: memoryThreshold withInvocation: actMemory];
		else if (cycle >= activatorThresholdDelay)
			fibroProduction |= [colony checkDoubleThreshold: activatorThreshold forGrid: activator];
		
		if (fibroProduction) {
			// produce fibronectin
			[colony productionForGrid: fibronectin];
			
			// check if cells get stuck on fibronectin
			[colony checkStuckOnFibronectin: fibronectin];
			
			// round up of cells on fibronectin
			[colony cellDoubleRoundUp: fibronectinThreshold forGrid: fibronectin];
			
			// once fibronectin production has started
			// determine if we should stop the reaction-diffusion
			if (disableCellActivator) {
				double max = [grid maxValueForDoubleGrid: @"activator"];
				if (max < disableCellActivatorThreshold)
					performRD = NO;
			}
		}
#endif
	}
	
	// dump final matrices
	[grid dumpGrid: @"cells" toFile: fp_cell];
	[grid dumpGrid: @"activator" toFile: fp_morpha];
	[grid dumpGrid: @"inhibitor" toFile: fp_morphb];
	[grid dumpGrid: @"fibronectin" toFile: fp_FN];
	//[grid dumpGrid: @"cells" toFile: fp_FNcell];
	[grid dumpGrid: @"cells" toFile: fp_FNcell withInvocation: actMemory];
	
	// close output files
	fclose(fp_cell);
	fclose(fp_morpha);
	fclose(fp_morphb);
	fclose(fp_FN);
	fclose(fp_FNcell);

#if 0
	if (enableMatlab)
		callMatlab(fibronectin);
#endif

	return;
}

@end
