//
//  SignalTransductionModel.m
//  LimbChondrogenesis
//
//  Created by Scott Christley on 7/10/06.
//  Copyright 2006 __MyCompanyName__. All rights reserved.
//

#import "SignalTransductionModel.h"
#import "MultiScale2DGrid.h"
#import "CellColony.h"
#import "Grid2DArray.h"
#import "Diffusion.h"
#import "SignalTransductionCell.h"

#import "internal.h"

@implementation SignalTransductionModel

- (void)doSim: (int)run
{
	// file pointer
	FILE *fp_cell, *fp_morpha, *fp_morphb, *fp_FN, *fp_params;
	FILE *fp_msga, *fp_msgb;
	NSString *s;
	
	// output file name prefix
	NSMutableString *fileName = [NSMutableString string];
	[fileName appendString: [simParameters objectForKey: @"outputName"]];
	[fileName appendFormat: @"_%d", run];
	
	// open the output files
	NSString *outputFile = [NSString stringWithFormat: @"%@_cell.dat", fileName];
	fp_cell = fopen([outputFile cString], "wb+");
	if (fp_cell == NULL) {
		printf("cannot open file\n");
		exit(0);
	}
	
	outputFile = [NSString stringWithFormat: @"%@_morpha.dat", fileName];
	fp_morpha = fopen([outputFile cString], "wb+");
	if (fp_morpha == NULL) {
		printf("cannot open file\n");
		exit(0);
	}
	
	outputFile = [NSString stringWithFormat: @"%@_morphb.dat", fileName];
	fp_morphb = fopen([outputFile cString], "wb+");
	if (fp_morphb == NULL) {
		printf("cannot open file\n");
		exit(0);
	}

	outputFile = [NSString stringWithFormat: @"%@_msga.dat", fileName];
	fp_msga = fopen([outputFile cString], "wb+");
	if (fp_msga == NULL) {
		printf("cannot open file\n");
		exit(0);
	}
	
	outputFile = [NSString stringWithFormat: @"%@_msgb.dat", fileName];
	fp_msgb = fopen([outputFile cString], "wb+");
	if (fp_msgb == NULL) {
		printf("cannot open file\n");
		exit(0);
	}
	
	outputFile = [NSString stringWithFormat: @"%@_FN.dat", fileName];
	fp_FN = fopen([outputFile cString], "wb+");
	if (fp_FN == NULL) {
		printf("cannot open file\n");
		exit(0);
	}
	
	outputFile = [NSString stringWithFormat: @"%@_st_matlab.dat", fileName];
	fp_params = fopen([outputFile cString], "wb+");
	if (fp_params == NULL) {
		printf("cannot open file\n");
		exit(0);
	}
	
	// save parameter set
	outputFile = [NSString stringWithFormat: @"%@_params.dat", fileName];
	[simParameters writeToFile: outputFile atomically: NO];
	
	// Initialize RNG
	s = [simParameters objectForKey: @"randomSeed"];
	srand([s intValue]);
	
	// create spatial grid
	s = [simParameters objectForKey: @"width"];
	int width = [s intValue];
	s = [simParameters objectForKey: @"height"];
	int height = [s intValue];
	s = [simParameters objectForKey: @"ratio"];
	int ratio = [s intValue];
	s = [simParameters objectForKey: @"boundaryCondition"];
	int bCond = PERIODIC_BOUNDARY;
	if ([s isEqualToString: @"NoFlux"])
		bCond = NOFLUX_BOUNDARY;
	MultiScale2DGrid *grid = [[MultiScale2DGrid alloc] initWithWidth: width andHeight: height];
	id cells = [grid newGridWithName: @"cells" atScale: 1 withEncode: idEncode];
	[cells setObject: [[NSNumber alloc] initWithInt: bCond] forKey: @"boundaryCondition"];

	s = [simParameters objectForKey: @"diffuseInhibitorMessenger"];
	BOOL diffuseInhibitorMessenger = [s boolValue];

	// messenger activator and inhibitor diffuse
	id msgActivator = [grid newGridWithName: @"msgActivator" atScale: ratio withEncode: intEncode];
	[msgActivator setObject: [[NSNumber alloc] initWithInt: bCond] forKey: @"boundaryCondition"];
	id msgInhibitor = [grid newGridWithName: @"msgInhibitor" atScale: ratio withEncode: intEncode];
	[msgInhibitor setObject: [[NSNumber alloc] initWithInt: bCond] forKey: @"boundaryCondition"];

	// activator and inhibitor limited to cells
	//id activator = [grid newGridWithName: @"activator" atScale: 1 withEncode: intEncode];
	//[activator setObject: [[NSNumber alloc] initWithInt: bCond] forKey: @"boundaryCondition"];
	//id inhibitor = [grid newGridWithName: @"inhibitor" atScale: 1 withEncode: intEncode];
	//[inhibitor setObject: [[NSNumber alloc] initWithInt: bCond] forKey: @"boundaryCondition"];

	id fibronectin = [grid newGridWithName: @"fibronectin" atScale: ratio withEncode: intEncode];
	[fibronectin setObject: [[NSNumber alloc] initWithInt: bCond] forKey: @"boundaryCondition"];

	id junk = [grid newGridWithName: @"junk" atScale: ratio withEncode: intEncode];
	[junk setObject: [[NSNumber alloc] initWithInt: bCond] forKey: @"boundaryCondition"];
	
	// test diffusion
	//Grid2DArray *m = [msgActivator objectForKey: @"matrix"];
	//int (*activatorGrid)[height*ratio][width*ratio] = [m getMatrix];
	//(*activatorGrid)[60][60] = 10000;
	//m = [msgInhibitor objectForKey: @"matrix"];
	//int (*inhibitorGrid)[height*ratio][width*ratio] = [m getMatrix];
	//(*inhibitorGrid)[60][60] = 10000;

	// create cell colony
	s = [simParameters objectForKey: @"cells"];
	int numCells = [s intValue];
	id colony = [[CellColony alloc] initWithSpatialGrid: cells];
	NSArray *bc = [colony createColonyWithCellNumber: numCells ofClass: [SignalTransductionCell class]];
	//[colony createColonyWithCellNumber: numCells];

	BOOL useMemory = NO;
	s = [simParameters objectForKey: @"useMemory"];
	if (s)
		useMemory = [s boolValue];
	[colony recordActivatorMemory: NO];
	s = [simParameters objectForKey: @"memoryDelay"];
	int memoryDelay = [s intValue];
	s = [simParameters objectForKey: @"memoryActive"];
	int memoryActive = [s intValue];
	
	[colony setFibronectinGrid: fibronectin];
	//NSArray *bc = [colony createColonyWithCellNumber: 100];
	//printf("%d\n", [bc count]);
	
	// diffusion process
	id diffusionProcess = [Diffusion new];

	// initial uniform and random cell distribution
	[grid dumpGrid: @"cells" toFile: fp_cell];
	s = [simParameters objectForKey: @"initialCellDiffusionSteps"];
	if (s == nil)
		[colony performDiffusion: 100 withMoveProbability: 0.8];
	else
		[colony performDiffusion: [s intValue] withMoveProbability: 0.8];
	[grid dumpGrid: @"cells" toFile: fp_cell];
	
	// simulation parameters
	s = [simParameters objectForKey: @"cycles"];
	int totCycles = [s intValue];
	s = [simParameters objectForKey: @"cellMovement"];
	int cellMovement = [s intValue];
	s = [simParameters objectForKey: @"cellKeepShape"];
	double cellKeepShape = [s doubleValue];
	s = [simParameters objectForKey: @"cellRestProb"];
	double cellRestProb = [s doubleValue];
	s = [simParameters objectForKey: @"outputRate"];
	int outputRate = [s intValue];
	s = [simParameters objectForKey: @"activatorThreshold"];
	int activatorThreshold = [s intValue];
	s = [simParameters objectForKey: @"activatorThresholdDelay"];
	int activatorThresholdDelay = [s intValue];
	s = [simParameters objectForKey: @"fibronectinThreshold"];
	int fibronectinThreshold = [s intValue];
	s = [simParameters objectForKey: @"maxFibronectin"];
	int maxFibronectin = [s intValue];
	s = [simParameters objectForKey: @"fibronectinDelay"];
	int fibronectinDelay = [s intValue];
	s = [simParameters objectForKey: @"memoryThreshold"];
	int memoryThreshold = [s intValue];
	s = [simParameters objectForKey: @"fibronectinCellMovement"];
	int fibronectinCellMovement = [s intValue];
	s = [simParameters objectForKey: @"fibronectinCellKeepShape"];
	double fibronectinCellKeepShape = [s doubleValue];
	s = [simParameters objectForKey: @"fibronectinCellRestProb"];
	double fibronectinCellRestProb = [s doubleValue];
	s = [simParameters objectForKey: @"disableCellActivator"];
	BOOL disableCellActivator = [s boolValue];
	s = [simParameters objectForKey: @"disableCellActivatorThreshold"];
	int disableCellActivatorThreshold = [s intValue];
	s = [simParameters objectForKey: @"disableReactionDiffusion"];
	int disableReactionDiffusion = [s intValue];
	s = [simParameters objectForKey: @"enableMatlab"];
	BOOL enableMatlab = [s boolValue];

	NSMutableDictionary *msgaParams = [NSMutableDictionary new];
	[msgActivator setObject: msgaParams forKey: @"parameters"];
	[msgaParams setObject: [simParameters objectForKey: @"actMessengerDiffusion"] forKey: @"diffusionRate"];
	[msgaParams setObject: [simParameters objectForKey: @"actMessengerDiffusionTime"] forKey: @"diffusionTime"];

	NSMutableDictionary *msgbParams = [NSMutableDictionary new];
	[msgInhibitor setObject: msgbParams forKey: @"parameters"];
	[msgbParams setObject: [simParameters objectForKey: @"inhMessengerDiffusion"] forKey: @"diffusionRate"];
	[msgbParams setObject: [simParameters objectForKey: @"inhMessengerDiffusionTime"] forKey: @"diffusionTime"];

	// cache reaction parameters within each cell
	s = [simParameters objectForKey: @"activatorOnActivator"];
	double activatorOnActivator = [s doubleValue];
	s = [simParameters objectForKey: @"activatorOnInhibitor"];
	double activatorOnInhibitor = [s doubleValue];
	s = [simParameters objectForKey: @"inhibitorOnActivator"];
	double inhibitorOnActivator = [s doubleValue];
	s = [simParameters objectForKey: @"inhibitorOnInhibitor"];
	double inhibitorOnInhibitor = [s doubleValue];
	s = [simParameters objectForKey: @"cellActivator"];
	int basalActivator = [s intValue];
	s = [simParameters objectForKey: @"basalInhibitor"];
	int basalInhibitor = [s intValue];
	s = [simParameters objectForKey: @"activatorMembraneTransmit"];
	int activatorMembraneTransmit = [s intValue];
	s = [simParameters objectForKey: @"activatorMembraneReceive"];
	int activatorMembraneReceive = [s intValue];
	s = [simParameters objectForKey: @"inhibitorMembraneTransmit"];
	int inhibitorMembraneTransmit = [s intValue];
	s = [simParameters objectForKey: @"inhibitorMembraneReceive"];
	int inhibitorMembraneReceive = [s intValue];
	s = [simParameters objectForKey: @"maxActivatorProduce"];
	int maxActivatorProduce = [s intValue];
	s = [simParameters objectForKey: @"maxInhibitorProduce"];
	int maxInhibitorProduce = [s intValue];
	int i;
	for (i = 0; i < [bc count]; ++i) {
		SignalTransductionCell *c = [bc objectAtIndex: i];
		[c setBasalActivator: basalActivator];
		[c setActivatorOnActivator: activatorOnActivator];
		[c setActivatorOnInhibitor: activatorOnInhibitor];
		[c setMaxActivatorProduce: maxActivatorProduce];
		[c setBasalInhibitor: basalInhibitor];
		[c setInhibitorOnActivator: inhibitorOnActivator];
		[c setInhibitorOnInhibitor: inhibitorOnInhibitor];
		[c setMaxInhibitorProduce: maxInhibitorProduce];
		[c setActivatorMembraneTransmit: activatorMembraneTransmit];
		[c setActivatorMembraneReceive: activatorMembraneReceive];
		[c setInhibitorMembraneTransmit: inhibitorMembraneTransmit];
		[c setInhibitorMembraneReceive: inhibitorMembraneReceive];
	}

	// save parameters for Matlab
	fwrite(&width, sizeof(int), 1, fp_params);
	fwrite(&height, sizeof(int), 1, fp_params);
	fwrite(&ratio, sizeof(int), 1, fp_params);
	fwrite(&outputRate, sizeof(int), 1, fp_params);
	fwrite(&totCycles, sizeof(int), 1, fp_params);
	fwrite(&activatorThreshold, sizeof(int), 1, fp_params);
	fwrite(&fibronectinThreshold, sizeof(int), 1, fp_params);
	s = [simParameters objectForKey: @"activatorDiffusion"];
	double tmp1 = [s doubleValue];
	s = [simParameters objectForKey: @"activatorDiffusionTime"];
	double tmp2 = [s doubleValue];
	tmp1 = tmp1 * tmp2;
	fwrite(&tmp1, sizeof(double), 1, fp_params);
	s = [simParameters objectForKey: @"activatorOnActivator"];
	tmp1 = [s doubleValue];
	fwrite(&tmp1, sizeof(double), 1, fp_params);
	s = [simParameters objectForKey: @"activatorOnInhibitor"];
	tmp1 = [s doubleValue];
	fwrite(&tmp1, sizeof(double), 1, fp_params);
	s = [simParameters objectForKey: @"maxActivatorProduce"];
	int tmp3 = [s intValue];
	fwrite(&tmp3, sizeof(int), 1, fp_params);
	s = [simParameters objectForKey: @"cellActivator"];
	tmp3 = [s intValue];
	fwrite(&tmp3, sizeof(int), 1, fp_params);
	s = [simParameters objectForKey: @"inhibitorDiffusion"];
	tmp1 = [s doubleValue];
	s = [simParameters objectForKey: @"inhibitorDiffusionTime"];
	tmp2 = [s doubleValue];
	tmp1 = tmp1 * tmp2;
	fwrite(&tmp1, sizeof(double), 1, fp_params);
	s = [simParameters objectForKey: @"inhibitorOnActivator"];
	tmp1 = [s doubleValue];
	fwrite(&tmp1, sizeof(double), 1, fp_params);
	s = [simParameters objectForKey: @"inhibitorOnInhibitor"];
	tmp1 = [s doubleValue];
	fwrite(&tmp1, sizeof(double), 1, fp_params);
	s = [simParameters objectForKey: @"maxInhibitorProduce"];
	tmp3 = [s intValue];
	fwrite(&tmp3, sizeof(int), 1, fp_params);
	int itmp = (int)useMemory;
	fwrite(&itmp, sizeof(int), 1, fp_params);
	if (useMemory)
		fwrite(&memoryThreshold, sizeof(int), 1, fp_params);
	else
		fwrite(&activatorThreshold, sizeof(int), 1, fp_params);
	fwrite(&fibronectinThreshold, sizeof(int), 1, fp_params);
	fwrite(&maxFibronectin, sizeof(int), 1, fp_params);
	fwrite(&fibronectinDelay, sizeof(int), 1, fp_params);
	fclose(fp_params);
	
	// setup invocations
	NSMethodSignature *sig = [[BiologicalCell new] methodSignatureForSelector: @selector(stuckOnFibronectin)];
	[sig retain];
	NSInvocation *stuckInv = [NSInvocation invocationWithMethodSignature: sig];
	[stuckInv setSelector: @selector(stuckOnFibronectin)];
	[stuckInv retain];

	sig = [[SignalTransductionCell new] methodSignatureForSelector: @selector(activatorInCell)];
	[sig retain];
	NSInvocation *actInCell = [NSInvocation invocationWithMethodSignature: sig];
	[actInCell setSelector: @selector(activatorInCell)];
	[actInCell retain];

	sig = [[SignalTransductionCell new] methodSignatureForSelector: @selector(inhibitorInCell)];
	[sig retain];
	NSInvocation *inhInCell = [NSInvocation invocationWithMethodSignature: sig];
	[inhInCell setSelector: @selector(inhibitorInCell)];
	[inhInCell retain];

	sig = [[SignalTransductionCell new] methodSignatureForSelector: @selector(performReaction)];
	[sig retain];
	NSInvocation *pReaction = [NSInvocation invocationWithMethodSignature: sig];
	[pReaction setSelector: @selector(performReaction)];
	[pReaction retain];

	sig = [[SignalTransductionCell new] methodSignatureForSelector: @selector(performSignalTransduction)];
	[sig retain];
	NSInvocation *pSignal = [NSInvocation invocationWithMethodSignature: sig];
	[pSignal setSelector: @selector(performSignalTransduction)];
	[pSignal retain];
	
	// main simulation loop
	int cycle;
	BOOL fibroProduction = NO;
	BOOL performRD = YES;
	BOOL triggerMemory = YES;
	BOOL activeMemory = YES;
	for (cycle = 0; cycle < totCycles; cycle++) {
		if ((cycle % outputRate) == 0) printf("cycle=%d\n",cycle);
		//printf("COUNT_FNcells=%d\n",COUNT_FNcells);
		
		// data output
		if ((cycle % outputRate) == 0) {
			[grid dumpGrid: @"cells" toFile: fp_cell];
			//[grid dumpGrid: @"activator" toFile: fp_morpha];
			[grid dumpGrid: @"cells" toFile: fp_morpha withInvocation: actInCell];
			//[grid dumpGrid: @"inhibitor" toFile: fp_morphb];
			[grid dumpGrid: @"cells" toFile: fp_morphb withInvocation: inhInCell];
			[grid dumpGrid: @"msgActivator" toFile: fp_msga];
			[grid dumpGrid: @"msgInhibitor" toFile: fp_msgb];
			[grid dumpGrid: @"fibronectin" toFile: fp_FN];
		}

		// only test diffusion
		//[reactDiff performDiffusion];
		//[reactDiff performDoubleDiffusion];

		//
		if (triggerMemory) {
			if ((useMemory) && (cycle >= memoryDelay)) {
				[colony recordActivatorMemory: YES];
				triggerMemory = NO;
			}
		} else if (activeMemory) {
			if ((useMemory) && (cycle >= (memoryDelay + memoryActive))) {
				[colony recordActivatorMemory: NO];
				activeMemory = NO;
			}
		}
		
		// cell diffusion
		if ((cycle % cellMovement) == 0) {
			[colony performDiffusion: 1 withShapeProbability: cellKeepShape
				 withMoveProbability: (1 - cellRestProb)];
		}
		
		// cell diffusion on fibronectin
		if ((cycle % fibronectinCellMovement) == 0) {
			[colony performDiffusion: 1 withShapeProbability: cellKeepShape
				 withMoveProbability: (1 - cellRestProb) forCellInvocation: stuckInv];
		}

		// check if disable reaction diffusion
		if ((disableReactionDiffusion != 0) && (cycle > disableReactionDiffusion))
			performRD = NO;

		// molecular reaction
		[colony cellsPerform: NO withInvocation: pReaction];
		[colony cellsPerform: NO withInvocation: pSignal];

		// molecular diffusion
		[diffusionProcess performDiffusionForGrid: msgActivator];
		if (diffuseInhibitorMessenger)
			[diffusionProcess performDiffusionForGrid: msgInhibitor];

#if 0
		// fibronectin production
		if (useMemory)
			fibroProduction |= [colony checkThreshold: memoryThreshold withInvocation: actMemory];
		else if (cycle >= activatorThresholdDelay)
			fibroProduction |= [colony checkThreshold: activatorThreshold forGrid: activator];
		
		if (fibroProduction) {
			// produce fibronectin
			[colony productionForGrid: fibronectin];
			
			// check if cells get stuck on fibronectin
			[colony checkStuckOnFibronectin: fibronectin];
			
			// round up of cells on fibronectin
			[colony cellRoundUp: fibronectinThreshold forGrid: fibronectin];
			
			// once fibronectin production has started
			// determine if we should stop the reaction-diffusion
			if (disableCellActivator) {
				int max = [grid maxValueForIntGrid: @"activator"];
				if (max < disableCellActivatorThreshold)
					performRD = NO;
			}
		}
#endif
	}
	
	// dump final matrices
	[grid dumpGrid: @"cells" toFile: fp_cell];
	//[grid dumpGrid: @"activator" toFile: fp_morpha];
	[grid dumpGrid: @"cells" toFile: fp_morpha withInvocation: actInCell];
	//[grid dumpGrid: @"inhibitor" toFile: fp_morphb];
	[grid dumpGrid: @"cells" toFile: fp_morphb withInvocation: inhInCell];
	[grid dumpGrid: @"msgActivator" toFile: fp_msga];
	[grid dumpGrid: @"msgInhibitor" toFile: fp_msgb];
	[grid dumpGrid: @"fibronectin" toFile: fp_FN];
	
	// close output files
	fclose(fp_cell);
	fclose(fp_morpha);
	fclose(fp_morphb);
	fclose(fp_msga);
	fclose(fp_msgb);
	fclose(fp_FN);

#if 0	
	if (enableMatlab)
		callMatlab(fibronectin);
#endif

	return;
}

@end
