/*
 main.m
 
 Main Body of LimbChondrogenesis.app
 
 Copyright (C) 2005 Scott Christley
 
 Author: Scott Christley <schristl@nd.edu>
 Author: Xuelian Zhu
 Date: 2005
 */

#ifdef GNUSTEP
#include <Foundation/Foundation.h>
#else
#import <Cocoa/Cocoa.h>
#endif

#import "MultiScale2DGrid.h"
#import "CellColony.h"
#import "Grid2DArray.h"
#import "ReactionDiffusion.h"
#import "BiologicalCell.h"
#import "ReactionDiffusionModel.h"
#import "SignalTransductionModel.h"

#ifndef GNUSTEP
#import <MatlabEngine/MatlabEngine.h>
#import <MatlabEngine/matrix.h>
#endif

#import "internal.h"

#ifdef __APPLE__
@implementation NSString (LimbChondrogenesisAdditions)
- (BOOL)boolValue
{
	if ([self caseInsensitiveCompare: @"YES"] == NSOrderedSame)
    {
		return YES;
    }
	if ([self caseInsensitiveCompare: @"true"] == NSOrderedSame)
    {
		return YES;
    }
	return [self intValue] != 0 ? YES : NO;
}
@end
#endif

// encode strings
NSString *idEncode = nil;
NSString *intEncode = nil;
NSString *doubleEncode = nil;
NSString *longEncode = nil;
NSString *boolEncode = nil;
NSDictionary *simParameters = nil;

// Generate normally distributed random number
double normal_dist_rand(double mean, double sd)
{
  double fac, radius, v1, v2;
  double rd1Value, rd2Value;

  do {
    rd1Value = RAND_NUM;
    rd2Value = RAND_NUM;
    v1 = (2.0 * rd1Value) - 1.0;
    v2 = (2.0 * rd2Value) - 1.0;
    radius = v1*v1 + v2*v2; 
  } while (radius >= 1.0);
  fac = sqrt (-2.0 * log (radius) / radius);
  return v2 * fac * sd + mean;    // use fixed params
}


/*
 void testMatlab()
 {
	 int i, r;
	 id me = [MatlabEngine new];
	 
	 BOOL ok = [me openEngine];
	 if (!ok) printf("Cannot start Matlab\n");
	 
	 //Engine *ep = [me getEngine];
	 //if (!ep) printf("bad engine ptr\n");
	 
	 mxArray *T = NULL;
	 T = mxCreateDoubleMatrix(1, 10, mxREAL);
	 double *tt = mxGetPr(T);
	 for (i = 0; i < 10; ++i) {
		 *tt = (double)i;
		 ++tt;
	 }
	 r = [me putVariable: T withName: @"T"];
	 if (r != 0) printf("error 1\n");
	 
	 r = [me evalString: @"D = 2 * T;"];
	 if (r != 0) printf("error 2\n");
	 
	 mxArray *result = NULL;
	 if ((result = [me getVariable: @"D"]) == NULL)
		 printf("Could not get D variable\n");
	 else {
		 tt = mxGetPr(result);
		 for (i = 0; i < 10; ++i) {
			 printf("%f\n", *tt);
			 ++tt;
		 }
		 mxFree(result);
	 }
	 
	 [me closeEngine];
 }
 */


// matlab engine
id me = nil;

// Matlab integration is future work
#if 0

#ifndef GNUSTEP
int callMatlab(id grid)
{
	//id me = [MatlabEngine new];
	//BOOL ok = [me openEngine];
	int r;
	
	if (!me) {
		me = [MatlabEngine new];
		BOOL ok = [me openEngine];
		if (!ok) printf("Cannot start Matlab\n");
	}
	
	id fGrid = [grid objectForKey: @"matrix"];
	int num = [fGrid countClusters: NO];
    printf("num of clusters %d\n", num);
	
	NSNumber *n = [grid objectForKey: @"width"];
	int width = [n intValue];
	n = [grid objectForKey: @"height"];
	int height = [n intValue];
	int (*molGrid)[height][width] = [fGrid getMatrix];
	
	//mxArray *T = mxCreateDoubleMatrix(width, height, mxREAL);
	mxArray *T = [me createRealMatrix: width : height];
	[me copy: T fromIntMatrix: molGrid : width : height];
	
	/*
	 double *mv = [me matrixRealPtr: T];
	 //double *mv = mxGetPr(T);
	 int i, j;
	 for (j = 0; j < width; ++j) {
		 for (i = 0; i < height; ++i) {
			 *mv = (double)(*molGrid)[i][j];
			 //printf("%d ", (*molGrid)[i][j]);
			 ++mv;
		 }
		 //printf("\n");
	 }
	 */
	
	r = [me putVariable: T withName: @"molecularGrid"];
	if (r != 0) printf("error 1\n");
	
	[me evalString: @"global ais"];
	
	NSString *s = [NSString stringWithFormat: @"API(molecularGrid, %d);", num];
	printf("evaluating: %s\n", [s cString]);
	r = [me evalString: s];
	if (r != 0) printf("error 2\n");
	
	printf("ais = %f\n", [me doubleValue: @"ais"]);
	printf("ans = %f\n", [me doubleValue: @"ans"]);
	/*
	 mxArray *result = NULL;
	 if ((result = [me getVariable: @"ais"]) == NULL)
	 printf("Could not get ais variable\n");
	 else {
		 double *tt = mxGetPr(result);
		 printf("ais = %f\n", *tt);
		 mxFree(result);
	 }
	 if ((result = [me getVariable: @"ans"]) == NULL)
	 printf("Could not get ans variable\n");
	 else {
		 double *tt = mxGetPr(result);
		 printf("api = %f\n", *tt);
		 mxFree(result);
	 }
	 */
	
	//[me closeEngine];
	
	return num;
}
#endif
#endif


int main (void)
{
	NSMutableDictionary *runParameters;
	NSAutoreleasePool *pool;
	NSString *paramFile;
	
	pool = [NSAutoreleasePool new];
	
	//testMatlab();
	
	// determine parameter file
	NSArray *args = [[NSProcessInfo processInfo] arguments];
	if ([args count] == 2)
		paramFile = [args objectAtIndex: 1];
	else
		paramFile = @"params.limb";
	
	// load parameter file
	runParameters = [[NSMutableDictionary dictionaryWithContentsOfFile: paramFile] retain];
	if (!runParameters) return 1;
	//printf("%s\n", [[runParameters description] cString]);
	
	// initialize encode strings
	idEncode = [[NSString alloc] initWithCString: @encode(id)];
	intEncode = [[NSString alloc] initWithCString: @encode(int)];
	longEncode = [[NSString alloc] initWithCString: @encode(long)];
	doubleEncode = [[NSString alloc] initWithCString: @encode(double)];
	boolEncode = [[NSString alloc] initWithCString: @encode(BOOL)];
	
	// add initial random seed if none
	NSString *s = [runParameters objectForKey: @"randomSeed"];
	if (s == nil)
		[runParameters setObject: START_RAND_SEED_STRING forKey: @"randomSeed"];
	
	// get model
	id model;
	s = [runParameters objectForKey: @"model"];
	if ([s isEqualToString: @"SignalTransduction"])
		model = [SignalTransductionModel new];
	else
		model = [ReactionDiffusionModel new];

	// perform simulation runs
	s = [runParameters objectForKey: @"runs"];
	if (s == nil) {
		int aRun;
		// single run
		simParameters = [[NSDictionary dictionaryWithDictionary: runParameters] retain];
		s = [simParameters objectForKey: @"outputRun"];
		if (s != nil)
			aRun = [s intValue];
		else
			aRun = 0;
		[model doSim: aRun];
		//doSim(aRun);
	} else {
		// multiple runs
		int i, cnt = [s intValue];
		[runParameters removeObjectForKey: @"runs"];
		
		for (i = 0; i < cnt; ++i) {
			printf("Starting run %d.\n", i);
			simParameters = [[NSDictionary dictionaryWithDictionary: runParameters] retain];
			[model doSim: i];
			//doSim(i);
			NSNumber *n = [NSNumber numberWithInt: rand()];
			[runParameters setObject: [n stringValue] forKey: @"randomSeed"];
		}
	}
	
	//runSimulation();
	if (me) [me closeEngine];
	
	return 0;
}
